
// HandCanvas.java
// Andrew Davison, dandrew@ratree.psu.ac.th, July 2004

/* Display a 3D model.
   The model can be moved left, right, up, down, forward, and back,
   and rotated left/right around the y-axis.

   Holding a key down causing key repetition, even in implementations
   which don't support keyRepeated().

   The model-specific data is located in the methods:

       getVerts()    : returns its vertices
       getNormals()  : returns its normals
       getTexCoords() : returns its texture coordinates
       getStripLengths() : returns an array containing the lengths of each 
                           of the triangle strips in the shape
       setMatColours()     : sets the material colours and shininess

       getColourCoords(): returns its colour coordinates

   These methods are generated by the ObjView Java 3D application; they
   can be copied over from the examObj.txt file that ObjView outputs.

   The texture image filename should be placed in the TEX_FNM constant.

   The camera's position may need to be adjusted, by changing the
   X_CAMERA, Y_CAMERA, Z_CAMERA values.
*/ 


import javax.microedition.lcdui.*;
import javax.microedition.m3g.*;


public class HandCanvas extends Canvas
                   implements CommandListener
{
  private static final String TEX_FNM = "/r.gif";
    // name of tex image file; change as required

  private static final float X_CAMERA = 0.0f; 
  private static final float Y_CAMERA = 0.0f; 
  private static final float Z_CAMERA = 3.0f; 
    // camera coordinates; change as required

  private static final float MODEL_MOVE = 0.2f; 
    // distance to move the model up/down/left/right/fwd/back for a key press

  private static final float MODEL_ROT = 5.0f;     // 5 degrees
    // angle to rotate the model left/right for a key press


  private HandDisplay top;

  private Graphics3D g3d;
  private Camera camera;
  private Light light;

  private Background bg;

  private VertexBuffer vertBuf; 
     // holds the model's positions, normals, colors coords, and tex coords
  private IndexBuffer idxBuf;   
     // indices to the model's triangle strips

  private Appearance app;

  private Transform camTrans, modelTrans;
    // transforms used by the camera and the model

  // model position and angle information
  private float xTrans, yTrans, zTrans;
  private int totalDegrees;

  private KeyRepeat keyRepeater;   // for repeating key presses


  public HandCanvas(HandDisplay top)
  {
    this.top = top;
    setCommandListener(this);
    keyRepeater = new KeyRepeat(this);
    keyRepeater.start();

    try {
      createScene();
    }
    catch (Exception e) 
    {  e.printStackTrace(); }
  } // end of HandCanvas()


  public void commandAction(Command c, Displayable d)
  {  
    if (c.getCommandType() == Command.EXIT) {
       keyRepeater.cancel();
       keyRepeater = null;
       top.quitApp();
     }
  } // end of commandAction()


  private void createScene() throws Exception
  {
    addCommand( new Command("Exit", Command.EXIT, 1) );

    g3d = Graphics3D.getInstance();

    // create a camera
    camera = new Camera();
    float aspectRatio = ((float) getWidth()) / ((float) getHeight());
    camera.setPerspective(45.0f, aspectRatio, 0.1f, 50.0f);
    // was 60, aspectRatio, 1, 1000

    // set up the camera's position
    camTrans = new Transform();
    camTrans.postTranslate(X_CAMERA, Y_CAMERA, Z_CAMERA);

    // create a light
    light = new Light();
    light.setColor(0xffffff);  // white light
    light.setIntensity(1.25f); // over bright

    // initialise the model's transform vars
    modelTrans = new Transform();
    xTrans = 0.0f; yTrans = 0.0f; zTrans = 0.0f;
    totalDegrees = 0;

    makeGeometry();
    makeAppearance();

    bg = new Background();
    bg.setColor(0x9EFEFE); // light blue 
  } // end of createScene()


  // ------------------- model creation -----------------------------

  private void makeGeometry()
  /* Read in the values for the vertices, normals, texture coords,
     and colour coords, and create a VertexBuffer object. Also
     read in the strip lengths array to create the index for the
     triangle strips used in the vertex buffer.

     Depending on the model, there may not be any texture or colour
     coordinates, and so the calls related to them should be commented
     out.

     The texture coordinates can use 2, 3, or 4 components, and the
     colour coordinates 3 or 4, which will require changes to the array
     creation code. The code here assumes (s,t) format for texture coords
     (2 components) and RGB for the colours (3 components), the most common 
     formats.
  */
  {
    // create vertices
    short[] verts = getVerts();
    VertexArray va = new VertexArray(verts.length/3, 3, 2);
    va.set(0, verts.length/3, verts);

    // create normals
    byte[] norms = getNormals();
    VertexArray normArray = new VertexArray(norms.length/3, 3, 1);
    normArray.set(0, norms.length/3, norms);
/*
    // create texture coordinates
    short[] tcs = getTexCoordsRev();
    VertexArray texArray = new VertexArray(tcs.length/2, 2, 2);
                          // this assumes (s,t) texture coordinates
    texArray.set(0, tcs.length/2, tcs);
*/
/*
    // create colour coordinates
    short[] cols = getColourCoords();
    VertexArray colsArray = new VertexArray(cols.length/3, 3, 2);
                          // this assumes RGB colour coordinates
    colsArray.set(0, cols.length/3, cols);
*/

    // ------ create the VertexBuffer for the model -------

    vertBuf = new VertexBuffer();
    float[] pbias = {(1.0f/255.0f), (1.0f/255.0f), (1.0f/255.0f)};
    vertBuf.setPositions(va, (2.0f/255.0f), pbias); // scale, bias
       // fix the scale and bias to create points in range [-1 to 1]

    vertBuf.setNormals(normArray);

    // vertBuf.setTexCoords(0, texArray, (1.0f/255.0f), null);
        // fix the scale to create texCoords in range [0 to 1] 

    // vertBuf.setColors(colsArray);

    // create the index buffer for the model (this tells MIDP how to
    // create triangle strips from the contents of the vertex buffer).
    idxBuf = new TriangleStripArray(0, getStripLengths() );
  }  // end of makeGeometry()

/*
  private short[] getTexCoordsRev()
  // Return an array of texture coords (s,t)
  // with their t coords reversed. 

  // This method will need to be commented out if there are 
  // no textures used in the model.
  { 
    short[] tcs = getTexCoords();

    // t' = 255 - t   (will later be scaled from 255 to 1)
    for(int i=1; i < tcs.length; i=i+2)
      tcs[i] = (short)(255 - tcs[i]);

    return tcs;
  }  // end of getTexCoordsRev()
*/


  private void makeAppearance() throws Exception
  /* Load the texture and material information into the Appearance
     node. 

     Depending on the model, there may not be a texture image, and so
     the texture-specific code here will need to be commented out.

     The material colour and shininess data is obtained from 
     setMatColours().
  */
  {
/*
    // load the image for the texture
    Image im = Image.createImage(TEX_FNM);

    // create an Image2D for the Texture2D
    Image2D image2D = new Image2D(Image2D.RGB, im);

    // create the Texture2D and enable mip mapping
    // the texture color is modulated with the lit material color
    Texture2D tex = new Texture2D(image2D);
    tex.setFiltering(Texture2D.FILTER_NEAREST, Texture2D.FILTER_NEAREST);
    tex.setWrapping(Texture2D.WRAP_CLAMP, Texture2D.WRAP_CLAMP);
    // tex.setWrapping(Texture2D.WRAP_REPEAT, Texture2D.WRAP_REPEAT);
    tex.setBlending(Texture2D.FUNC_MODULATE);
*/
    // create the appearance
    app = new Appearance();
    // app.setTexture(0, tex);
    app.setMaterial( setMatColours() );
  }  // end of makeAppearance()


  // ------------------ paint the canvas ------------------


  public void paint(Graphics g)
  {
    // bind the canvas graphic to our Graphics3D object
    g3d.bindTarget(g, true, Graphics3D.DITHER | Graphics3D.TRUE_COLOR);

    g3d.clear(bg);    // clear the color and depth buffers

    g3d.setCamera(camera, camTrans);   // position the camera

    // set up a "headlight": a directional light shining
    // from the direction of the camera
    g3d.resetLights();
    g3d.addLight(light, camTrans);

    updateModelTrans();

    // Render the model. We provide the vertex and index buffers
    // to specify the geometry; the appearance so we know what
    // material and texture to use; and a tranform to position
    // the model
    g3d.render(vertBuf, idxBuf, app, modelTrans);

    g3d.releaseTarget();      // flush
  } // end of paint()


  private void updateModelTrans()
  // the model's transform = a translation * a rotation
  {
    modelTrans.setIdentity();   // reset
    modelTrans.postTranslate(xTrans, yTrans, zTrans);
    modelTrans.postRotate(totalDegrees, 0, 1, 0);  // around y-axis
  }  // end of updateModelTrans()



  // -----------------key operation methods ---------------------


  public void keyRepeated(int keyCode)
  {
    int gameAction = getGameAction(keyCode);
    if (hasRepeatEvents())
      performAction( gameAction );
  }


  protected void keyPressed(int keyCode)
  {
    int gameAction = getGameAction(keyCode);

    // System.out.println("Can repeat: " + hasRepeatEvents());

    // if this MIDP doesn't support repeated keys 
    // then initialise our key repeater
    if (!hasRepeatEvents())
      keyRepeater.startRepeat( gameAction );
  } // end of keyPressed()


  protected void keyReleased(int keyCode)
  {
    int gameAction = getGameAction(keyCode);
    // if this MIDP doesn't support repeated keys 
    // then stop our key repeater
    if (!hasRepeatEvents())
     keyRepeater.stopRepeat( gameAction );
 }



  public void performAction(int gameAction) 
  // called by KeyRepeat object or the keyRepeated() method
  {
    if (gameAction == UP) {
      moveModel(0,MODEL_MOVE,0);  // y up
    }
    else if (gameAction == DOWN) {
      moveModel(0,-MODEL_MOVE,0);  // y down
    }
    else if (gameAction == LEFT) {
      moveModel(-MODEL_MOVE,0,0);   // x left
    }
    else if (gameAction == RIGHT) {
      moveModel(MODEL_MOVE,0,0);   // x right
    }
    else if (gameAction == GAME_A) {
      // System.out.println("fwd");
      moveModel(0,0,MODEL_MOVE);   // z fwd
    }
    else if (gameAction == GAME_B) {
      // System.out.println("back");
      moveModel(0,0,-MODEL_MOVE);   // z back
    }
    else if (gameAction == GAME_C) {
      // System.out.println("rot left");
      rotYModel(-MODEL_ROT);   // rotate left
    }
    else if (gameAction == GAME_D) {
      // System.out.println("rot right");
      rotYModel(MODEL_ROT);   // rotate right
    }
  } // end of keyPressed()


  private void moveModel(float x, float y, float z)
  // update the model's translation values
  { xTrans += x; yTrans += y; zTrans += z;
    repaint();
  }


  private void rotYModel(float degrees)
  // update the model's y-axis rotation value
  { totalDegrees += degrees;
    repaint();
  }  // end of rotYModel()


  // ----------------- data-specific methods added here -------
/* 
  The following methods can be pasted in from the ObjView application
  output stored in examObj.txt:

  private short[] getVerts() {...}
  private byte[] getNormals() {...}
  private short[] getTexCoords() (...}
  private short[] getColourCoords() (...}
  private int[] getStripLengths() {...}
  private Material setMatColours() {...}
*/


// Methods for model in hand.obj

  private short[] getVerts()
  // return an array holding Verts [2205 values / 3 = 735 points] 
  {
    short[] vals = {
      	42,114,-5,  	47,101,4,  	52,114,-2,  	47,101,4,  
	61,111,-1,  	47,101,4,  	57,99,4,  	50,83,9,  
	57,99,4,  	62,96,-2,  	61,111,-1,  	89,69,25,  
	87,72,30,  	97,68,28,  	96,71,35,  	101,62,33,  
	83,69,38,  	90,67,42,  	87,72,30,  	90,67,42,  
	96,71,35,  	95,63,46,  	96,71,35,  	100,66,39,  
	101,62,33,  	100,66,39,  	101,56,38,  	100,66,39,  
	100,58,43,  	95,63,46,  	100,58,43,  	97,59,47,  
	92,51,49,  	97,59,47,  	87,55,49,  	95,63,46,  
	87,55,49,  	90,67,42,  	82,58,46,  	83,69,38,  
	77,61,42,  	83,69,38,  	77,69,34,  	87,72,30,  
	79,70,28,  	87,72,30,  	82,69,24,  	89,69,25,  
	82,69,24,  	85,62,26,  	82,69,24,  	78,63,26,  
	79,70,28,  	74,64,30,  	77,69,34,  	73,64,36,  
	77,61,42,  	69,56,37,  	77,61,42,  	70,52,42,  
	82,58,46,  	70,52,42,  	74,48,46,  	70,52,42,  
	64,43,40,  	70,52,42,  	64,46,35,  	69,56,37,  
	70,57,32,  	73,64,36,  	70,57,32,  	74,64,30,  
	70,57,32,  	78,63,26,  	74,55,28,  	85,62,26,  
	79,51,27,  	85,62,26,  	91,58,29,  	89,69,25,  
	91,58,29,  	97,68,28,  	91,58,29,  	101,62,33,  
	95,53,34,  	101,56,38,  	95,49,40,  	100,58,43,  
	94,49,45,  	92,51,49,  	87,40,46,  	92,51,49,  
	84,42,49,  	87,55,49,  	79,45,49,  	82,58,46,  
	79,45,49,  	74,48,46,  	71,37,46,  	74,48,46,  
	67,40,43,  	64,43,40,  	67,40,43,  	59,36,38,  
	67,40,43,  	61,33,42,  	67,40,43,  	65,31,44,  
	71,37,46,  	69,29,44,  	71,37,46,  	76,35,46,  
	79,45,49,  	76,35,46,  	84,42,49,  	79,33,43,  
	87,40,46,  	88,40,41,  	94,49,45,  	88,40,41,  
	95,49,40,  	87,42,35,  	95,53,34,  	84,46,30,  
	91,58,29,  	84,46,30,  	79,51,27,  	75,41,26,  
	79,51,27,  	70,45,26,  	74,55,28,  	70,45,26,  
	70,57,32,  	66,47,30,  	64,46,35,  	59,38,33,  
	64,43,40,  	59,38,33,  	59,36,38,  	53,31,34,  
	59,36,38,  	53,29,39,  	61,33,42,  	55,27,43,  
	65,31,44,  	59,24,46,  	65,31,44,  	64,22,45,  
	69,29,44,  	69,20,42,  	69,29,44,  	74,27,40,  
	76,35,46,  	74,27,40,  	79,33,43,  	81,32,38,  
	88,40,41,  	81,34,33,  	87,42,35,  	78,37,29,  
	84,46,30,  	78,37,29,  	75,41,26,  	71,33,24,  
	75,41,26,  	66,36,23,  	70,45,26,  	66,36,23,  
	66,47,30,  	62,38,27,  	59,38,33,  	56,31,27,  
	53,31,34,  	51,31,26,  	44,26,35,  	41,39,27,  
	30,29,29,  	41,39,27,  	31,51,20,  	41,55,19,  
	31,51,20,  	35,68,14,  	27,65,9,  	32,83,4,  
	27,65,9,  	26,64,-1,  	21,54,1,  	26,64,-1,  
	29,61,-8,  	34,79,-12,  	29,61,-8,  	39,61,-8,  
	36,45,-4,  	45,51,2,  	46,35,11,  	47,41,19,  
	51,31,26,  	47,41,19,  	41,39,27,  	47,41,19,  
	41,55,19,  	47,41,19,  	46,55,11,  	45,51,2,  
	49,67,6,  	45,51,2,  	47,64,-3,  	39,61,-8,  
	43,77,-12,  	34,79,-12,  	39,97,-16,  	34,79,-12,  
	30,81,-5,  	26,64,-1,  	30,81,-5,  	32,83,4,  
	37,102,-1,  	32,83,4,  	41,84,10,  	35,68,14,  
	44,69,14,  	41,55,19,  	44,69,14,  	46,55,11,  
	44,69,14,  	49,67,6,  	54,80,3,  	49,67,6,  
	52,78,-6,  	47,64,-3,  	52,78,-6,  	43,77,-12,  
	49,95,-16,  	39,97,-16,  	43,110,-17,  	39,97,-16,  
	34,100,-9,  	30,81,-5,  	34,100,-9,  	37,102,-1,  
	42,114,-5,  	37,102,-1,  	47,101,4,  	41,84,10,  
	50,83,9,  	44,69,14,  	50,83,9,  	54,80,3,  
	62,96,-2,  	52,78,-6,  	59,94,-10,  	49,95,-16,  
	59,94,-10,  	63,107,-13,  	62,96,-2,  	64,111,-7,  
	61,111,-1,  	64,111,-7,  	54,117,-11,  	63,107,-13,  
	54,110,-17,  	49,95,-16,  	54,110,-17,  	43,110,-17,  
	54,117,-11,  	43,110,-17,  	42,114,-12,  	34,100,-9,  
	42,114,-12,  	42,114,-5,  	54,117,-11,  	52,114,-2,  
	61,111,-1,  	53,31,34,  	44,26,35,  	53,29,39,  
	45,20,42,  	55,27,43,  	49,17,46,  	59,24,46,  
	53,14,48,  	59,24,46,  	58,12,48,  	64,22,45,  
	64,9,44,  	69,20,42,  	68,8,38,  	69,20,42,  
	73,19,36,  	74,27,40,  	77,26,35,  	81,32,38,  
	77,26,35,  	81,34,33,  	77,27,31,  	81,34,33,  
	75,29,27,  	78,37,29,  	75,29,27,  	71,33,24,  
	66,25,23,  	66,36,23,  	61,28,23,  	62,38,27,  
	61,28,23,  	56,31,27,  	52,23,19,  	51,31,26,  
	52,23,19,  	46,35,11,  	37,24,4,  	36,45,-4,  
	23,44,-6,  	29,61,-8,  	23,44,-6,  	21,54,1,  
	17,64,-11,  	18,67,-3,  	15,86,-17,  	18,67,-3,  
	18,89,-9,  	14,91,-1,  	14,112,-7,  	3,111,-3,  
	13,124,-11,  	3,111,-3,  	3,126,-10,  	-7,122,-10,  
	3,126,-10,  	3,127,-19,  	13,124,-11,  	15,125,-19,  
	13,124,-11,  	18,110,-15,  	14,112,-7,  	18,110,-15,  
	18,89,-9,  	15,107,-22,  	15,86,-17,  	15,107,-22,  
	5,106,-24,  	14,121,-24,  	4,121,-25,  	14,121,-24,  
	3,127,-19,  	14,121,-24,  	15,125,-19,  	14,121,-24,  
	18,110,-15,  	15,107,-22,  	-21,90,15,  	-27,101,22,  
	-22,91,24,  	-29,90,31,  	-22,91,24,  	-17,76,25,  
	-21,90,15,  	-16,75,16,  	-21,90,15,  	-26,88,11,  
	-21,90,15,  	-28,101,16,  	-27,101,22,  	-38,102,22,  
	-27,101,22,  	-34,101,28,  	-29,90,31,  	-34,101,28,  
	-42,97,32,  	-38,102,22,  	-42,97,32,  	-46,97,28,  
	-44,85,30,  	-46,97,28,  	-47,94,22,  	-38,102,22,  
	-40,97,16,  	-28,101,16,  	-30,98,11,  	-26,88,11,  
	-30,98,11,  	-36,85,14,  	-40,97,16,  	-36,85,14,  
	-47,94,22,  	-44,84,22,  	-44,85,30,  	-37,70,29,  
	-44,85,30,  	-38,87,34,  	-42,97,32,  	-38,87,34,  
	-29,90,31,  	-38,87,34,  	-31,72,34,  	-37,70,29,  
	-29,52,28,  	-37,70,29,  	-37,68,21,  	-44,84,22,  
	-37,68,21,  	-36,85,14,  	-30,69,13,  	-26,88,11,  
	-21,72,11,  	-16,75,16,  	-21,72,11,  	-10,59,14,  
	-14,55,8,  	-5,51,10,  	-4,39,2,  	-5,51,10,  
	-1,62,-7,  	-4,65,2,  	-7,86,-6,  	-4,89,2,  
	-7,86,-6,  	-10,107,-12,  	-7,86,-6,  	-3,84,-15,  
	-1,62,-7,  	8,61,-12,  	-1,62,-7,  	10,36,-6,  
	-4,39,2,  	10,36,-6,  	12,5,1,  	24,13,0,  
	26,-12,7,  	40,6,11,  	45,-5,17,  	58,14,22,  
	63,10,26,  	71,22,27,  	63,10,26,  	74,20,31,  
	63,10,26,  	68,8,31,  	57,-10,30,  	60,-8,37,  
	52,-22,36,  	60,-8,37,  	47,-22,41,  	56,-7,44,  
	47,-22,41,  	50,-4,48,  	41,-20,46,  	44,-2,49,  
	35,-20,47,  	44,-2,49,  	31,-17,46,  	40,0,47,  
	31,-17,46,  	35,2,43,  	25,-18,43,  	35,2,43,  
	20,-11,39,  	31,10,38,  	20,-11,39,  	19,17,29,  
	10,12,31,  	19,17,29,  	8,43,21,  	20,49,15,  
	6,68,12,  	15,69,7,  	6,68,12,  	5,91,4,  
	-4,89,2,  	-7,109,-5,  	-10,107,-12,  	-7,109,-5,  
	-7,122,-10,  	-7,109,-5,  	3,111,-3,  	5,91,4,  
	14,91,-1,  	15,69,7,  	18,67,-3,  	15,69,7,  
	21,54,1,  	20,49,15,  	27,65,9,  	20,49,15,  
	31,51,20,  	20,49,15,  	30,29,29,  	19,17,29,  
	30,29,29,  	31,10,38,  	44,26,35,  	31,10,38,  
	45,20,42,  	35,2,43,  	49,17,46,  	40,0,47,  
	53,14,48,  	44,-2,49,  	53,14,48,  	50,-4,48,  
	58,12,48,  	56,-7,44,  	64,9,44,  	56,-7,44,  
	68,8,38,  	60,-8,37,  	68,8,38,  	68,8,31,  
	73,19,36,  	74,20,31,  	77,26,35,  	74,20,31,  
	77,27,31,  	71,22,27,  	75,29,27,  	71,22,27,  
	66,25,23,  	58,14,22,  	61,28,23,  	58,14,22,  
	52,23,19,  	40,6,11,  	37,24,4,  	24,13,0,  
	23,44,-6,  	10,36,-6,  	23,44,-6,  	8,61,-12,  
	17,64,-11,  	8,61,-12,  	6,84,-19,  	-3,84,-15,  
	-6,105,-20,  	-10,107,-12,  	-6,119,-23,  	-10,107,-12,  
	-8,122,-17,  	-7,122,-10,  	-8,122,-17,  	3,127,-19,  
	-6,119,-23,  	4,121,-25,  	-6,119,-23,  	5,106,-24,  
	-6,105,-20,  	5,106,-24,  	6,84,-19,  	15,86,-17,  
	17,64,-11,  	12,-123,45,  	12,-47,17,  	14,-127,44,  
	16,-49,17,  	17,-128,43,  	20,-49,19,  	20,-128,44,  
	20,-49,22,  	21,-127,47,  	20,-49,22,  	20,-125,49,  
	18,-46,26,  	17,-121,49,  	15,-43,27,  	17,-121,49,  
	12,-44,27,  	15,-122,49,  	10,-47,26,  	12,-123,45,  
	10,-46,20,  	12,-47,17,  	5,-31,14,  	12,-47,17,  
	19,-38,13,  	16,-49,17,  	19,-38,13,  	20,-49,19,  
	33,-35,20,  	20,-49,22,  	30,-40,30,  	18,-46,26,  
	24,-35,40,  	15,-43,27,  	15,-31,40,  	12,-44,27,  
	6,-22,39,  	12,-44,27,  	0,-32,37,  	10,-47,26,  
	-2,-35,33,  	10,-47,26,  	-1,-38,24,  	10,-46,20,  
	-1,-38,24,  	5,-31,14,  	-1,-38,24,  	-8,-27,28,  
	-2,-35,33,  	-5,-25,37,  	0,-32,37,  	-3,-17,40,  
	6,-22,39,  	2,11,33,  	6,-22,39,  	13,-18,39,  
	15,-31,40,  	20,-26,42,  	24,-35,40,  	27,-27,44,  
	24,-35,40,  	32,-29,44,  	24,-35,40,  	38,-32,36,  
	30,-40,30,  	46,-27,29,  	33,-35,20,  	50,-10,22,  
	33,-35,20,  	27,-27,12,  	19,-38,13,  	17,-24,8,  
	5,-31,14,  	0,2,6,  	5,-31,14,  	-8,-20,19,  
	-8,-27,28,  	-14,-13,24,  	-8,-27,28,  	-13,-14,32,  
	-5,-25,37,  	-9,-13,39,  	-3,-17,40,  	-6,11,37,  
	2,11,33,  	-10,38,32,  	-3,45,23,  	-17,58,32,  
	-11,60,24,  	-17,58,32,  	-23,75,32,  	-17,58,32,  
	-24,55,33,  	-10,38,32,  	-17,35,33,  	-6,11,37,  
	-13,10,37,  	-9,-13,39,  	-18,8,31,  	-13,-14,32,  
	-18,8,31,  	-14,-13,24,  	-18,5,22,  	-14,-13,24,  
	-11,3,14,  	-8,-20,19,  	-11,3,14,  	0,2,6,  
	-11,3,14,  	-16,31,11,  	-18,5,22,  	-23,30,20,  
	-18,8,31,  	-23,32,29,  	-13,10,37,  	-23,32,29,  
	-17,35,33,  	-23,32,29,  	-24,55,33,  	-29,52,28,  
	-24,55,33,  	-31,72,34,  	-23,75,32,  	-29,90,31,  
	-23,75,32,  	-17,76,25,  	-11,60,24,  	-16,75,16,  
	-11,60,24,  	-10,59,14,  	-3,45,23,  	-5,51,10,  
	-2,67,11,  	-4,65,2,  	-2,67,11,  	-4,89,2,  
	-2,67,11,  	6,68,12,  	-3,45,23,  	8,43,21,  
	2,11,33,  	10,12,31,  	13,-18,39,  	20,-11,39,  
	20,-26,42,  	25,-18,43,  	27,-27,44,  	31,-17,46,  
	27,-27,44,  	35,-20,47,  	32,-29,44,  	41,-20,46,  
	38,-32,36,  	47,-22,41,  	46,-27,29,  	52,-22,36,  
	46,-27,29,  	57,-10,30,  	50,-10,22,  	63,10,26,  
	50,-10,22,  	45,-5,17,  	27,-27,12,  	26,-12,7,  
	17,-24,8,  	12,5,1,  	0,2,6,  	-4,39,2,  
	-16,31,11,  	-23,52,12,  	-16,31,11,  	-29,51,20,  
	-23,30,20,  	-29,51,20,  	-23,32,29,  	-29,51,20,  
	-29,52,28,  	-29,51,20,  	-37,68,21,  	-23,52,12,  
	-30,69,13,  	-23,52,12,  	-21,72,11,  	-23,52,12,  
	-14,55,8,  	-23,52,12,  	-4,39,2      };
    return vals;
  }  // end of getVerts()


  private byte[] getNormals()
  // return an array holding Normals [2205 values / 3 = 735 points] 
  {
    byte[] vals = {
      	-69,95,51,  	-17,51,116,  	-4,100,79,  	-17,51,116,  
	66,68,87,  	-17,51,116,  	62,17,111,  	61,8,112,  
	62,17,111,  	122,-29,28,  	66,68,87,  	33,41,-117,  
	-3,127,-8,  	79,58,-82,  	64,109,19,  	114,-3,-59,  
	-33,101,72,  	-10,95,85,  	-3,127,-8,  	-10,95,85,  
	64,109,19,  	47,73,94,  	64,109,19,  	114,48,34,  
	114,-3,-59,  	114,48,34,  	120,-43,-14,  	114,48,34,  
	112,-16,59,  	47,73,94,  	112,-16,59,  	64,23,109,  
	62,-31,108,  	64,23,109,  	-9,33,123,  	47,73,94,  
	-9,33,123,  	-10,95,85,  	-43,53,109,  	-33,101,72,  
	-65,70,85,  	-33,101,72,  	-68,105,29,  	-3,127,-8,  
	-74,94,-46,  	-3,127,-8,  	-34,61,-107,  	33,41,-117,  
	-34,61,-107,  	8,-25,-125,  	-34,61,-107,  	-69,18,-106,  
	-74,94,-46,  	-102,63,-45,  	-68,105,29,  	-103,73,25,  
	-65,70,85,  	-107,60,37,  	-65,70,85,  	-84,46,85,  
	-43,53,109,  	-84,46,85,  	-62,27,108,  	-84,46,85,  
	-95,49,71,  	-84,46,85,  	-113,60,10,  	-107,60,37,  
	-101,50,-61,  	-103,73,25,  	-101,50,-61,  	-102,63,-45,  
	-101,50,-61,  	-69,18,-106,  	-39,11,-122,  	8,-25,-125,  
	24,-13,-125,  	8,-25,-125,  	51,-42,-110,  	33,41,-117,  
	51,-42,-110,  	79,58,-82,  	51,-42,-110,  	114,-3,-59,  
	78,-64,-80,  	120,-43,-14,  	101,-78,-15,  	112,-16,59,  
	97,-67,50,  	62,-31,108,  	77,-79,66,  	62,-31,108,  
	22,-44,118,  	-9,33,123,  	-32,5,124,  	-43,53,109,  
	-32,5,124,  	-62,27,108,  	-37,-5,123,  	-62,27,108,  
	-69,32,103,  	-95,49,71,  	-69,32,103,  	-87,71,62,  
	-69,32,103,  	-60,57,98,  	-69,32,103,  	-25,25,123,  
	-37,-5,123,  	28,-29,121,  	-37,-5,123,  	12,-56,115,  
	-32,5,124,  	12,-56,115,  	22,-44,118,  	67,-87,65,  
	77,-79,66,  	98,-83,-2,  	97,-67,50,  	98,-83,-2,  
	101,-78,-15,  	89,-60,-70,  	78,-64,-80,  	64,-38,-104,  
	51,-42,-110,  	64,-38,-104,  	24,-13,-125,  	46,4,-120,  
	24,-13,-125,  	-28,36,-120,  	-39,11,-122,  	-28,36,-120,  
	-101,50,-61,  	-85,60,-75,  	-113,60,10,  	-103,75,-15,  
	-95,49,71,  	-103,75,-15,  	-87,71,62,  	-72,106,9,  
	-87,71,62,  	-68,83,70,  	-60,57,98,  	-49,63,100,  
	-25,25,123,  	-11,39,122,  	-25,25,123,  	42,3,121,  
	28,-29,121,  	83,-29,93,  	28,-29,121,  	75,-65,81,  
	12,-56,115,  	75,-65,81,  	67,-87,65,  	97,-82,19,  
	98,-83,-2,  	106,-51,-51,  	89,-60,-70,  	83,-22,-95,  
	64,-38,-104,  	83,-22,-95,  	46,4,-120,  	51,-3,-118,  
	46,4,-120,  	-17,32,-123,  	-28,36,-120,  	-17,32,-123,  
	-85,60,-75,  	-76,61,-84,  	-103,75,-15,  	-50,94,-71,  
	-72,106,9,  	64,111,5,  	-42,88,83,  	30,64,107,  
	-39,51,111,  	30,64,107,  	-35,52,111,  	54,43,108,  
	-35,52,111,  	-36,47,114,  	-100,55,57,  	-103,44,62,  
	-100,55,57,  	-120,33,-29,  	17,119,-45,  	-120,33,-29,  
	-58,0,-114,  	-63,-8,-111,  	-58,0,-114,  	43,-38,-115,  
	56,-35,-110,  	116,-26,-48,  	115,0,-57,  	119,35,32,  
	64,111,5,  	119,35,32,  	30,64,107,  	119,35,32,  
	54,43,108,  	119,35,32,  	122,5,39,  	116,-26,-48,  
	123,-31,16,  	116,-26,-48,  	101,-46,-65,  	43,-38,-115,  
	34,-38,-117,  	-63,-8,-111,  	-60,-4,-113,  	-63,-8,-111,  
	-123,24,-26,  	-120,33,-29,  	-123,24,-26,  	-103,44,62,  
	-96,50,69,  	-103,44,62,  	-30,43,117,  	-36,47,114,  
	71,12,106,  	54,43,108,  	71,12,106,  	122,5,39,  
	71,12,106,  	123,-31,16,  	115,-38,41,  	123,-31,16,  
	101,-55,-56,  	101,-46,-65,  	101,-55,-56,  	34,-38,-117,  
	32,-30,-120,  	-60,-4,-113,  	-40,46,-113,  	-60,-4,-113,  
	-117,41,-30,  	-123,24,-26,  	-117,41,-30,  	-96,50,69,  
	-69,95,51,  	-96,50,69,  	-17,51,116,  	-30,43,117,  
	61,8,112,  	71,12,106,  	61,8,112,  	115,-38,41,  
	122,-29,28,  	101,-55,-56,  	94,-49,-72,  	32,-30,-120,  
	94,-49,-72,  	96,8,-84,  	122,-29,28,  	116,54,1,  
	66,68,87,  	116,54,1,  	26,121,-33,  	96,8,-84,  
	31,37,-119,  	32,-30,-120,  	31,37,-119,  	-40,46,-113,  
	26,121,-33,  	-40,46,-113,  	-74,96,-40,  	-117,41,-30,  
	-74,96,-40,  	-69,95,51,  	26,121,-33,  	-4,100,79,  
	66,68,87,  	-72,106,9,  	-42,88,83,  	-68,83,70,  
	-66,63,90,  	-49,63,100,  	-52,49,106,  	-11,39,122,  
	-17,23,125,  	-11,39,122,  	39,-4,122,  	42,3,121,  
	80,-26,97,  	83,-29,93,  	109,-49,46,  	83,-29,93,  
	112,-52,35,  	75,-65,81,  	111,-63,14,  	97,-82,19,  
	111,-63,14,  	106,-51,-51,  	105,-50,-53,  	106,-51,-51,  
	82,-28,-94,  	83,-22,-95,  	82,-28,-94,  	51,-3,-118,  
	37,-15,-122,  	-17,32,-123,  	-12,37,-122,  	-76,61,-84,  
	-12,37,-122,  	-50,94,-71,  	68,33,-103,  	64,111,5,  
	68,33,-103,  	115,0,-57,  	69,-28,-104,  	56,-35,-110,  
	25,-15,-125,  	-58,0,-114,  	25,-15,-125,  	17,119,-45,  
	91,-7,-90,  	127,14,4,  	88,-20,-91,  	127,14,4,  
	127,2,5,  	90,29,86,  	92,30,84,  	12,47,119,  
	68,82,71,  	12,47,119,  	-4,99,82,  	-78,76,68,  
	-4,99,82,  	-14,122,-35,  	68,82,71,  	94,85,-19,  
	68,82,71,  	127,16,1,  	92,30,84,  	127,16,1,  
	127,2,5,  	93,-15,-87,  	88,-20,-91,  	93,-15,-87,  
	-1,-22,-126,  	66,39,-103,  	-12,42,-121,  	66,39,-103,  
	-14,122,-35,  	66,39,-103,  	94,85,-19,  	66,39,-103,  
	127,16,1,  	93,-15,-87,  	113,50,-34,  	77,95,38,  
	107,48,51,  	56,40,108,  	107,48,51,  	111,40,50,  
	113,50,-34,  	112,41,-46,  	113,50,-34,  	36,19,-121,  
	113,50,-34,  	44,100,-67,  	77,95,38,  	-39,122,-10,  
	77,95,38,  	23,100,77,  	56,40,108,  	23,100,77,  
	-35,64,105,  	-39,122,-10,  	-35,64,105,  	-106,59,41,  
	-107,-27,65,  	-106,59,41,  	-117,18,-50,  	-39,122,-10,  
	-62,64,-92,  	44,100,-67,  	-13,48,-118,  	36,19,-121,  
	-13,48,-118,  	-69,-17,-107,  	-62,64,-92,  	-69,-17,-107,  
	-117,18,-50,  	-114,-46,-35,  	-107,-27,65,  	-101,-49,62,  
	-107,-27,65,  	-32,-4,124,  	-35,64,105,  	-32,-4,124,  
	56,40,108,  	-32,-4,124,  	-23,-12,125,  	-101,-49,62,  
	-104,-44,61,  	-101,-49,62,  	-109,-46,-49,  	-114,-46,-35,  
	-109,-46,-49,  	-69,-17,-107,  	-60,-19,-112,  	36,19,-121,  
	28,19,-124,  	112,41,-46,  	28,19,-124,  	103,63,-42,  
	21,48,-117,  	-88,78,-51,  	-63,-5,-111,  	-88,78,-51,  
	-104,-27,-70,  	-128,-4,9,  	-127,-18,-7,  	-92,20,86,  
	-127,-18,-7,  	-128,-4,-5,  	-127,-18,-7,  	-91,-35,-84,  
	-104,-27,-70,  	-18,-35,-122,  	-104,-27,-70,  	-28,-32,-121,  
	-63,-5,-111,  	-28,-32,-121,  	-21,-29,-123,  	34,-28,-120,  
	44,-31,-116,  	70,-21,-105,  	70,-28,-103,  	58,-15,-113,  
	80,-38,-92,  	75,-41,-96,  	80,-38,-92,  	105,-56,-46,  
	80,-38,-92,  	108,-55,-40,  	99,-56,-60,  	109,-63,24,  
	97,-84,10,  	109,-63,24,  	71,-77,74,  	78,-46,90,  
	71,-77,74,  	35,-26,121,  	34,-55,111,  	-17,2,127,  
	-11,-25,125,  	-17,2,127,  	-50,12,117,  	-53,29,113,  
	-50,12,117,  	-62,37,106,  	-54,16,115,  	-62,37,106,  
	-41,31,117,  	-56,49,104,  	-41,31,117,  	-19,46,118,  
	6,37,122,  	-19,46,118,  	25,44,118,  	-4,66,110,  
	15,50,117,  	94,42,76,  	15,50,117,  	14,41,120,  
	-92,20,86,  	-73,31,101,  	-128,-4,-5,  	-73,31,101,  
	-78,76,68,  	-73,31,101,  	12,47,119,  	14,41,120,  
	90,29,86,  	94,42,76,  	127,14,4,  	94,42,76,  
	17,119,-45,  	-4,66,110,  	-100,55,57,  	-4,66,110,  
	-35,52,111,  	-4,66,110,  	-39,51,111,  	-19,46,118,  
	-39,51,111,  	-56,49,104,  	-42,88,83,  	-56,49,104,  
	-66,63,90,  	-62,37,106,  	-52,49,106,  	-53,29,113,  
	-17,23,125,  	-17,2,127,  	-17,23,125,  	35,-26,121,  
	39,-4,122,  	78,-46,90,  	80,-26,97,  	78,-46,90,  
	109,-49,46,  	109,-63,24,  	109,-49,46,  	108,-55,-40,  
	112,-52,35,  	105,-56,-46,  	111,-63,14,  	105,-56,-46,  
	105,-50,-53,  	75,-41,-96,  	82,-28,-94,  	75,-41,-96,  
	37,-15,-122,  	58,-15,-113,  	-12,37,-122,  	58,-15,-113,  
	68,33,-103,  	70,-21,-105,  	69,-28,-104,  	34,-28,-120,  
	25,-15,-125,  	-28,-32,-121,  	25,-15,-125,  	-18,-35,-122,  
	91,-7,-90,  	-18,-35,-122,  	-8,-34,-123,  	-91,-35,-84,  
	-77,-29,-98,  	-128,-4,-5,  	-78,25,-98,  	-128,-4,-5,  
	-106,67,-25,  	-78,76,68,  	-106,67,-25,  	-14,122,-35,  
	-78,25,-98,  	-12,42,-121,  	-78,25,-98,  	-1,-22,-126,  
	-77,-29,-98,  	-1,-22,-126,  	-8,-34,-123,  	88,-20,-91,  
	91,-7,-90,  	-127,-9,-11,  	-74,-38,-98,  	-64,-36,-105,  
	2,-41,-121,  	22,-39,-120,  	81,-49,-87,  	111,-19,-61,  
	119,-43,18,  	127,5,11,  	119,-43,18,  	98,25,78,  
	49,-43,110,  	14,35,122,  	3,-48,119,  	14,35,122,  
	-24,-34,121,  	-88,21,90,  	-92,-59,67,  	-127,-9,-11,  
	-108,-53,-43,  	-74,-38,-98,  	-70,-49,-96,  	-74,-38,-98,  
	17,-46,-119,  	2,-41,-121,  	17,-46,-119,  	81,-49,-87,  
	83,-72,-66,  	119,-43,18,  	64,-105,36,  	49,-43,110,  
	-3,-81,99,  	3,-48,119,  	-24,-50,115,  	-24,-34,121,  
	-5,-15,127,  	-24,-34,121,  	-15,-59,113,  	-92,-59,67,  
	-86,-80,51,  	-92,-59,67,  	-97,-78,-29,  	-108,-53,-43,  
	-97,-78,-29,  	-70,-49,-96,  	-97,-78,-29,  	-112,-62,-12,  
	-86,-80,51,  	-84,-50,83,  	-15,-59,113,  	-1,-5,127,  
	-5,-15,127,  	34,31,119,  	-5,-15,127,  	-15,19,126,  
	-24,-50,115,  	-36,-5,123,  	-3,-81,99,  	-33,-27,121,  
	-3,-81,99,  	21,-72,104,  	-3,-81,99,  	60,-95,61,  
	64,-105,36,  	89,-90,-22,  	83,-72,-66,  	80,-43,-90,  
	83,-72,-66,  	56,-41,-108,  	17,-46,-119,  	-6,-38,-122,  
	-70,-49,-96,  	-58,-29,-111,  	-70,-49,-96,  	-88,-50,-79,  
	-112,-62,-12,  	-118,-40,-30,  	-112,-62,-12,  	-116,-41,36,  
	-84,-50,83,  	-69,-14,107,  	-1,-5,127,  	15,19,126,  
	34,31,119,  	45,24,117,  	56,61,98,  	65,24,107,  
	110,45,47,  	65,24,107,  	61,23,110,  	65,24,107,  
	-20,-9,126,  	45,24,117,  	-29,5,125,  	15,19,126,  
	-63,2,112,  	-69,-14,107,  	-120,-21,39,  	-116,-41,36,  
	-120,-21,39,  	-118,-40,-30,  	-117,-27,-45,  	-118,-40,-30,  
	-88,-29,-89,  	-88,-50,-79,  	-88,-29,-89,  	-58,-29,-111,  
	-88,-29,-89,  	-86,-24,-92,  	-117,-27,-45,  	-121,-31,-27,  
	-120,-21,39,  	-109,-25,63,  	-63,2,112,  	-109,-25,63,  
	-29,5,125,  	-109,-25,63,  	-20,-9,126,  	-104,-44,61,  
	-20,-9,126,  	-23,-12,125,  	61,23,110,  	56,40,108,  
	61,23,110,  	111,40,50,  	110,45,47,  	112,41,-46,  
	110,45,47,  	103,63,-42,  	56,61,98,  	-88,78,-51,  
	-97,36,76,  	-128,-4,9,  	-97,36,76,  	-92,20,86,  
	-97,36,76,  	15,50,117,  	56,61,98,  	25,44,118,  
	34,31,119,  	6,37,122,  	-15,19,126,  	-41,31,117,  
	-36,-5,123,  	-54,16,115,  	-33,-27,121,  	-50,12,117,  
	-33,-27,121,  	-11,-25,125,  	21,-72,104,  	34,-55,111,  
	60,-95,61,  	71,-77,74,  	89,-90,-22,  	97,-84,10,  
	89,-90,-22,  	99,-56,-60,  	80,-43,-90,  	80,-38,-92,  
	80,-43,-90,  	70,-28,-103,  	56,-41,-108,  	44,-31,-116,  
	-6,-38,-122,  	-21,-29,-123,  	-58,-29,-111,  	-63,-5,-111,  
	-86,-24,-92,  	-72,-16,-105,  	-86,-24,-92,  	-116,-40,-37,  
	-121,-31,-27,  	-116,-40,-37,  	-109,-25,63,  	-116,-40,-37,  
	-104,-44,61,  	-116,-40,-37,  	-109,-46,-49,  	-72,-16,-105,  
	-60,-19,-112,  	-72,-16,-105,  	28,19,-124,  	-72,-16,-105,  
	21,48,-117,  	-72,-16,-105,  	-63,-5,-111      };
    return vals;
  }  // end of getNormals()


  private int[] getStripLengths()
  // return an array holding the lengths of each triangle strip
  {
    int[] lens = {
      11, 5, 257, 77, 199, 186
    };
    return lens;
  }  // end of getStripLengths()


  private Material setMatColours()
  // set the material's colour and shininess values
  {
    Material mat = new Material();

    mat.setColor(Material.AMBIENT, 0x00765D4D);
    mat.setColor(Material.EMISSIVE, 0x00000000);
    mat.setColor(Material.DIFFUSE, 0xFF876A56);
    mat.setColor(Material.SPECULAR, 0x004D4D4D);
    mat.setShininess(60.0f);

    return mat;
  } // end of setMatColours()


} // end of HandCanvas class
