
// WrapObjView.java
// Andrew Davison, July 2004, dandrew@ratree.psu.ac.th

/* Load a model stored in a Wavefront file (.obj file).
   The model is triangulated and stripified as it is loaded. 
   The loaded model is displayed on-screen, and can be translated
   and rotated by using the mouse.

   The .obj file is translated into a BranchGroup with perhaps many
   Shape3D children. Each Shape3D child has a single Geometry node and
   Appearance node. (However, the code can deal with multiple geometries 
   below a Shape3D.)

   Because of the triangulation and stripification, the Geometry is
   a TriangleStripArray node, and its points are interleaved.
   Each 'point' is a tuple of values:
       textures coords, colour coords, normals, vertices

   A texture coord may have 2, 3, or 4 components.
   A colour coord may have 3 or 4 components.
   A normal has 3 components. A vertex has 3 components.

   An .obj file in Java 3D can't store colour coordinates,
   but this code still checks for them.

   All the .obj files with textures that I've seen only use 2D 
   texture coordinates (i.e. (s,t) values).

   Some .obj files only have material data (in a .mtl file),
   which is stored as colour and shininess values in the shape's 
   Appearance node.

   Java 3D understands a large set of usemtl names (e.g. flesh)
   without requiring a mtllib call in the .obj file. You can use this
   to quickly change the colour of a model, by adding a usemtl line
   to the file (e.g. usemtl flesh). The predefined names are listed 
   in Java 3D's ObjectFile documentation.

   OvjView writes output to the screen, reporting what it has found, and
   generates at most 6 methods for each Shape3D it detects in the model. 
   These methods are written to EXAMINE_FN (examObj.txt), to be pasted into 
   ModelCanvas.java used in the ModelDisplay MIDP application.

   The model methods for each Shape3D:
       getVerts()    : returns its vertices
       getNormals()  : returns its normals
       getTexCoords() : returns its texture coordinates
       getStripLengths() : returns an array containing the lengths of each 
                           of the triangle strips in the shape
       setMatColours()     : sets the material colours and shininess

       getColourCoords(): returns its colour coordinates
       (This will never be generated for a Java 3D .obj file.)

   If there are multiple Shape3Ds for the model, then the method names end
   with an integer: 1, 2, ... for each shape.

   ---------
   The main drawback with this approach is the generation of methods
   with very large arrays. The arrays may be too big for Java to compile.

   If a large array is split into pieces, distributed over several methods, 
   and constructed at run time, then a compile time error can be
   avoided. However, the MIDP emulator only allows an array to be at most 
   32 Kb large, so will crash when it has to render it.

   There doesn't seem any solution to this limit, so I just issue a warning
   and generate the methods anyway.

   ------------- CHANGES ---------------------

   21st October 2004
     In printAppearance(): switch on texture modulation if a Material
     and Texture are detected together
*/


import javax.swing.*;
import java.awt.*;
import java.util.*;
import java.io.*;

import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.geometry.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.behaviors.vp.*;

import com.sun.j3d.loaders.objectfile.ObjectFile;
import com.sun.j3d.loaders.*;



public class WrapObjView extends JPanel
{
  private static final int PWIDTH = 512;   // size of panel
  private static final int PHEIGHT = 512; 

  private static final int BOUNDSIZE = 100;  // larger than world

  private static final Point3d USERPOSN = new Point3d(0,0,5);
    // initial user position

  private static final String EXAMINE_FN = "examObj.txt";
    // the file where the model methods are written

  // private static final double CREASE_ANGLE = 60.0;

  private static final double MAX_ELEMS = 4000;
      /* (Probable) max number of elements in a coords array before 
         the MIDP 32kb byte[] array limit is reached. Needs testing. */


  private SimpleUniverse su;
  private BranchGroup sceneBG;
  private BoundingSphere bounds;   // for environment nodes

  private FileWriter ofw;     // for writing out model info

  // vertex format information
  private boolean isInterleaved;
  private boolean hasNormals;
  private int numColours;
  private int numTextures;
  private int numVertElems;
  private int pointsCount;

  // for numbering multiple copies of the model methods
  private boolean hasMultipleShapes = false;
  private int shapeCount;


  public WrapObjView(String fnm)
  // A panel holding a 3D canvas
  {
    setLayout( new BorderLayout() );
    setOpaque( false );
    setPreferredSize( new Dimension(PWIDTH, PHEIGHT));

    GraphicsConfiguration config =
					SimpleUniverse.getPreferredConfiguration();
    Canvas3D canvas3D = new Canvas3D(config);
    add("Center", canvas3D);
    canvas3D.setFocusable(true);     // give focus to the canvas 
    canvas3D.requestFocus();

    su = new SimpleUniverse(canvas3D);

    createSceneGraph(fnm);
    initUserPosition();        // set user's viewpoint
    orbitControls(canvas3D);   // controls for moving the viewpoint
    
    su.addBranchGraph( sceneBG );
  } // end of WrapObjView()



  private void createSceneGraph(String fnm) 
  // initilise the scene
  { 
    sceneBG = new BranchGroup();
    bounds = new BoundingSphere(new Point3d(0,0,0), BOUNDSIZE);   

    lightScene();         // add the lights
    addBackground();      // add the sky

    // load, examine, and display the model
    BranchGroup modelBG = loadModel(fnm);
    printGraphInfo(modelBG, fnm);
    sceneBG.addChild(modelBG);

    sceneBG.compile();   // fix the scene
  } // end of createSceneGraph()


  private void lightScene()
  /* One ambient light, 2 directional lights */
  {
    Color3f white = new Color3f(1.0f, 1.0f, 1.0f);

    // Set up the ambient light
    AmbientLight ambientLightNode = new AmbientLight(white);
    ambientLightNode.setInfluencingBounds(bounds);
    sceneBG.addChild(ambientLightNode);

    // Set up the directional lights
    Vector3f light1Direction  = new Vector3f(-1.0f, -1.0f, -1.0f);
       // left, down, backwards 
    Vector3f light2Direction  = new Vector3f(1.0f, -1.0f, 1.0f);
       // right, down, forwards

    DirectionalLight light1 = 
            new DirectionalLight(white, light1Direction);
    light1.setInfluencingBounds(bounds);
    sceneBG.addChild(light1);

    DirectionalLight light2 = 
        new DirectionalLight(white, light2Direction);
    light2.setInfluencingBounds(bounds);
    sceneBG.addChild(light2);
  }  // end of lightScene()



  private void addBackground()
  // A blue sky
  { Background back = new Background();
    back.setApplicationBounds( bounds );
    back.setColor(0.17f, 0.65f, 0.92f);    // sky colour
    sceneBG.addChild( back );
  }  // end of addBackground()



  private void orbitControls(Canvas3D c)
  /* OrbitBehaviour allows the user to rotate around the scene, and to
     zoom in and out.  */
  {
    OrbitBehavior orbit = 
		new OrbitBehavior(c, OrbitBehavior.REVERSE_ALL);
    orbit.setSchedulingBounds(bounds);

    ViewingPlatform vp = su.getViewingPlatform();
    vp.setViewPlatformBehavior(orbit);	 
  }  // end of orbitControls()



  private void initUserPosition()
  // Set the user's initial viewpoint using lookAt()
  {
    ViewingPlatform vp = su.getViewingPlatform();
    TransformGroup steerTG = vp.getViewPlatformTransform();

    Transform3D t3d = new Transform3D();
    steerTG.getTransform(t3d);

    // args are: viewer posn, where looking, up direction
    t3d.lookAt( USERPOSN, new Point3d(0,0,0), new Vector3d(0,1,0));
    t3d.invert();

    steerTG.setTransform(t3d);
  }  // end of initUserPosition()


  private BranchGroup loadModel(String fnm)
  // the loaded model is triangulated and stripified
  {
	int flags = ObjectFile.TRIANGULATE | ObjectFile.STRIPIFY;
	ObjectFile f = new ObjectFile(flags);
	                    // (float)(CREASE_ANGLE * Math.PI/180.0));
	Scene scene = null;
	try {
	  scene = f.load(fnm);
	}
	catch (FileNotFoundException e) {
	  System.err.println(e);
	  System.exit(1);
	}
	catch (ParsingErrorException e) {
	  System.err.println(e);
	  System.exit(1);
	}
	catch (IncorrectFormatException e) {
	  System.err.println(e);
	  System.exit(1);
	}
    return scene.getSceneGroup();
  }  // end of loadModel()



  // ---------------------- examine the loaded model -----------------


  private void printGraphInfo(BranchGroup bg, String fnm)
  /* Traverse the model's scene graph and store 
     information in EXAMINE_FN via the ofw FileWriter.
  */
  {
    System.out.println("Writing " + fnm + " model details to " + EXAMINE_FN);
    try {
      ofw = new FileWriter( EXAMINE_FN );
      fileWrite("\n// Methods for model in " + fnm  + "\n\n");
      examineNode(bg);
      ofw.close();
    }
    catch( IOException ioe )
    { System.err.println("Cannot write to " + EXAMINE_FN); }
  }  // end of printGraphInfo()


  private void examineNode(Node node)
  /* A Node can be a Group or a Leaf.
     Recursively call examineNode() on the children of a Group.
  */
  {
    if(node instanceof Group) {     // the Node is a Group
      Group g = (Group) node;
      // System.out.println("Group: " + g.getClass());

      if(g instanceof TransformGroup) {    // consider subclass
        Transform3D t3d = new Transform3D();
	    ((TransformGroup) g).getTransform(t3d);
        System.out.println(t3d.toString());   // show Transform3D info for TG
      }

      if (g.numChildren() > 1) {
        System.out.println(g.numChildren() + " children");
        fileWrite("// WARNING: created " + g.numChildren() + 
                                      " copies of all methods\n\n");
        hasMultipleShapes = true;
      }

      Enumeration enumKids = g.getAllChildren();
      shapeCount = 1;
      while(enumKids.hasMoreElements())  {   // visit Group children
        if (hasMultipleShapes) {
          fileWrite("// ============================ method set " + shapeCount + 
                    " =======================\n\n");
          System.out.println("================================");
          System.out.println("Child " + shapeCount);
          System.out.println("");
        }
        examineNode((Node) enumKids.nextElement());
        shapeCount++;
      }
    }
    else if (node instanceof Leaf) {     // the Node is a Leaf
      // System.out.println("Leaf: " + node.getClass());
      if (node instanceof Shape3D)
        examineShape3D((Shape3D) node);
    }
    // else    // the Node is something other than a Group or Leaf
      // System.out.println("Node: " + node.getClass());

  }  // end of examineNode()


  private void examineShape3D(Shape3D shape)
  /* A Shape3D is a container for Geometry and Appearance nodes.
     A shape may contain many geometries: examine each one with
     examineGeometry().
     Report appearance info with printAppearance().
  */
  {
    int numGeoms = shape.numGeometries();    // consider geometries
    if (numGeoms == 0)
      System.out.println("No Geometry Components");
    else if (numGeoms == 1) {
      Geometry g = shape.getGeometry();
      examineGeometry(1, g);
    }
    else {   // more than one geometry in the shape
      System.out.println("Num. of Geometries: " + numGeoms);
      fileWrite("// WARNING: created " + numGeoms + 
                                 " copies of all geometry methods\n\n");

      Enumeration enumGeoms = shape.getAllGeometries();
      int i = 1;
      while(enumGeoms.hasMoreElements()) {
        fileWrite("// ======================geometry method group " + i + 
                    " =======================\n\n");
        examineGeometry(i, (Geometry) enumGeoms.nextElement() );
        i++;
      }
    }
    System.out.println("");

    Appearance app = shape.getAppearance();    // consider appearance
    if (app == null)
      System.out.println("No Appearance Component");
    else
      printAppearance(app);
  }  // end of examineShape3D()



  // -------------------- report on geometry ----------

  private void examineGeometry(int index, Geometry geo)
  /* Due to triangulation and stripification, the geometry 
     will be a TriangleStripArray. Report on its interleaved
     coordinates, and the strip lengths. 
  */
  {
    System.out.println("Geometry: " + geo.getClass());
    if(geo instanceof TriangleStripArray) {
      TriangleStripArray tsa = (TriangleStripArray) geo;

      examineVertexFormat(tsa);

      pointsCount = tsa.getVertexCount();
      System.out.println("Total num. of Points: " + pointsCount); 

      if (isInterleaved)
        extractCoords(tsa);

      stripInfo(tsa);
    }
  }  // end of examineGeometry()



  private void examineVertexFormat(TriangleStripArray tsa)
  { 
    // reset vertex format globals
    isInterleaved = false;
    hasNormals = false;
    numColours = 0;
    numTextures = 0;
    numVertElems = 3;    // for (x,y,z) coord

    int vf = tsa.getVertexFormat();

    if ((vf & (GeometryArray.BY_REFERENCE)) != 0)
      System.out.println("VF: by reference");

    if ((vf & (GeometryArray.INTERLEAVED)) != 0) {
      isInterleaved = true;
      System.out.println("VF: interleaved");
    }
    else
      System.out.println("VF: **NOT** interleaved");

    // a colour coordinate may have 3 or 4 components
    if ((vf & GeometryArray.COLOR_3) != 0) {
      numColours = 3;
      System.out.println("VF: per vertex colours (3)");
    }
    else if ((vf & GeometryArray.COLOR_4) != 0) {
      numColours = 4;
      System.out.println("VF: per vertex colours (4)");
    }
    numVertElems += numColours;

    if ((vf & GeometryArray.NORMALS) != 0) {
      hasNormals = true;
      System.out.println("VF: per vertex normals");
      numVertElems += 3;
    }

    // a texture coordinate may have 2, 3 or 4 components
    if ((vf & GeometryArray.TEXTURE_COORDINATE_2) != 0) {
      numTextures = 2;
      System.out.println("VF: per vertex texture coords (2)");
    }
    else if ((vf & GeometryArray.TEXTURE_COORDINATE_3) != 0) {
      numTextures = 3;
      System.out.println("VF: per vertex texture coords (3)");
    }
    else if ((vf & GeometryArray.TEXTURE_COORDINATE_4) != 0) {
      numTextures = 4;
      System.out.println("VF: per vertex texture coords (4)");
    }
    numVertElems += numTextures;

    System.out.println("VF: num. of vertex elements: " + numVertElems);
  }  // end of examineVertexFormat()



  private void extractCoords(TriangleStripArray tsa)
  /* Report data for vertices, normals, texture coords, and colour coords.
     Each 'point' in the TriangleStrip array is a tuple of values:
       textures coords, colour coords, normals, vertices

     The number of textured coords and colour coords may vary. There's
     always 3 normals values, 3 vertex values.
  */
  {
    float[] itlvVerts = tsa.getInterleavedVertices();
    System.out.println("Num. of interleaved Verts: " + itlvVerts.length);

    int pointOffset = numTextures + numColours;
    if (hasNormals)
      pointOffset += 3;  // for the 3 normal values
    getCoords(itlvVerts, "Verts", pointOffset, 3);    // (x,y,z)

    if (hasNormals) {
      int normOffset = numTextures + numColours;
      getCoords(itlvVerts, "Normals", normOffset, 3);  // (x,y,z)
    }

    if (numTextures != 0)    // 2, 3, or 4 values
      getCoords(itlvVerts, "TexCoords", 0, numTextures);

    if (numColours != 0)     // 3 or 4 values
      getCoords(itlvVerts, "ColourCoords", numTextures, numColours);
  }  // end of extractCoords()


  private void getCoords(float[] iVerts, String coordType, 
                                        int offset, int numVals)
  /* Pull out the relevant coordinates from the TriangleStrip array,
     storing them in their own float array. Scale them up, converting
     them to integers. These are placed in an integer array, which becomes
     the values listed in the model methods.
  */
  {
    float[] coords = new float[pointsCount*numVals];   
    int posn = 0;

    // pull out the relevant numbers from the interleaved verts array
    for(int i=offset; i < iVerts.length; i += numVertElems) {
      for (int j=0; j < numVals; j++)
        coords[posn+j] = iVerts[i+j];
      posn += numVals;
    }

    // find the min and max values in the coords array
    float minVal = coords[0];
    float maxVal = coords[0];
    for(int i=1; i < coords.length; i ++) {
      if (minVal > coords[i])
        minVal = coords[i];
      if (maxVal < coords[i])
        maxVal = coords[i];
    }
    System.out.println("");
    System.out.println(coordType + " coords min/max: " + minVal + " / " + maxVal);

    printIntCoords(coords, coordType, minVal, maxVal, numVals);
  }  // end of getCoords()



  private void printIntCoords(float[] coords, String coordType, 
                               float minVal, float maxVal, int numVals)
  // convert floats to integers (by rescaling), and print them
  {
    int[] icoords = new int[coords.length];   

    // vertices and normals are scaled to integers in the range -128 to 127
    if ( coordType.equals("Verts") || coordType.equals("Normals") ) {
      System.out.println("Rescale " + coordType + " coords to [-128 to 127]");
      for(int i=0; i < coords.length; i ++) {
        icoords[i] = (int) Math.round( ((coords[i] - minVal) * 
                             (256.0f/(maxVal-minVal))) - 128.0f);
        if (icoords[i] == 128)
          icoords[i] = 127;
      }
      printCoords(icoords, coordType, numVals);
    }
    else if ( coordType.equals("ColourCoords") || coordType.equals("TexCoords") ) {
       // colour and texture coords are scaled to integers in the range 0 to 255
      System.out.println("Rescale " + coordType + " coords to [0 to 255]");
      for(int i=0; i < coords.length; i ++) {
        icoords[i] = (int) Math.round( (coords[i] - minVal) * 
                             (256.0f/(maxVal-minVal)) );
        if (icoords[i] == 256)
          icoords[i] = 255;
      }
      printCoords(icoords, coordType, numVals);
    }
  }  // end of printIntCoords()



  private void printCoords(int[] icoords, String coordType, int numVals)
  /* Print the integer values in icoords[] into the relevant model
     method written to examObj.txt
       * vertices --> getVerts()
       * normals --> getNormals()
       * texture coords --> getTexCoords()
       * colour coords --> getColourCoords()

     If there are multiple shapes in the model, then a number is included
     at the end of the model method names.
  */
  { int coordsLen = icoords.length;

    String arrayTypeStr = (coordType.equals("Normals")) ? "byte" : "short";
       // normals are returned in an array of bytes, the others are in short[]

    String methodName;
    if (hasMultipleShapes)
      methodName = "get" + coordType + shapeCount + "()";
    else
      methodName = "get" + coordType + "()";

    if (coordsLen > MAX_ELEMS)
      System.out.println("WARNING: " + coordType + " array very long: " +
                                                       coordsLen + " values");
      /* The method will probably not compile. Even if it's split into parts
         the emulator cannot handle arrays bigger than 32 Kb. */


    fileWrite("  private " + arrayTypeStr + "[] " + methodName + "\n" );
    fileWrite("  // return an array holding " + coordType + 
                  " [" +   coordsLen + " values / " + numVals +  " = " + 
                           (coordsLen/numVals) + " points] \n");
    if (coordsLen > MAX_ELEMS)
      fileWrite("  // WARNING: array very long\n");
    fileWrite("  {\n" );
    fileWrite("    " + arrayTypeStr + "[] vals = {\n      " );

    int coordCount = 0;
    for(int i=0; i < coordsLen; i += numVals) {
      printCoord(icoords, i, numVals, coordsLen); 
      coordCount++;
      if (coordCount == 4) {   // 4 coords per line
        coordCount = 0;
        fileWrite("\n");
      }
    }
    fileWrite("    };\n");
    fileWrite("    return vals;\n" );
    fileWrite("  }  // end of " + methodName + "\n\n\n" );
  }  // end of printCoords()



  private void printCoord(int[] iVerts, int posn, int numVals, int coordsLen)
  // print the multiple values making up a single coordinate
  {
    fileWrite("\t");
    for (int i=0; i < numVals; i++) {
      if ((posn+i) == (coordsLen-1))   // if the last point
        fileWrite("" + iVerts[posn+i]);  // then don't print a comma
      else
        fileWrite(iVerts[posn+i] + ",");
    }
    fileWrite("  ");
  } // end of printCoord()



  private void stripInfo(TriangleStripArray tsa)
  /* Output a list of strip lengths which states how the data in the
     interleaved array is split into distinct triangle strips.

     The data is embedded in the model method getStripLengths()
     written to examObj.txt

     If there are multiple shapes in the model, then a number is included
     at the end of the method name.
  */
  {
    int numStrips = tsa.getNumStrips();
    System.out.println("Num. strips: " + numStrips);

    int[] svCounts = new int[numStrips];
    tsa.getStripVertexCounts(svCounts);

    String methodName;
    if (hasMultipleShapes)
      methodName = "getStripLengths" + shapeCount + "()";
    else
      methodName = "getStripLengths()";

    fileWrite("  private int[] " + methodName + "\n" );
    fileWrite("  // return an array holding the lengths of each triangle strip\n" );
    fileWrite("  {\n" );
    fileWrite("    int[] lens = {\n      " );

    int stripsCount = 0;
    for(int i=0; i < numStrips-1; i++) {
      fileWrite(svCounts[i] + ", ");
      stripsCount++;
      if (stripsCount == 20) {   // 20 numbers per line
        stripsCount = 0;
        fileWrite("\n      ");
      }
    }
    fileWrite(svCounts[numStrips-1] + "\n    };\n");
    fileWrite("    return lens;\n" );
    fileWrite("  }  // end of " + methodName + "\n\n\n" );
  } // end of stripInfo()



  // -------------------- report on appearance (colour, shininess) ----------

  private void printAppearance(Appearance app)
  /* Report on the colour attributes, material, texture, and
     texture attributes (if they exist). The material and texture
     details are written to examObj.txt

     Change (21/Oct/04): switch on texture modulation if a Material
     and Texture are detected together
  */
  {
     ColoringAttributes ca = app.getColoringAttributes();
     if (ca != null)
       System.out.println(ca.toString());
     else
       System.out.println("No colouring attributes");

     Material mat = app.getMaterial();
     if (mat != null) {
       System.out.println(mat.toString());
       printMaterial(mat);
     }
     else
       System.out.println("No material");

     Texture tex = app.getTexture();
     if (tex != null) {
       ImageComponent[] ims = tex.getImages();
       System.out.println("Num. texture images: " + ims.length);
       for (int i=0; i < ims.length; i++)
         System.out.print("(" + ims[i].getWidth() + "," + ims[i].getHeight() + ") ");
       System.out.println("");
     }
     else
       System.out.println("No texture");

     if (tex != null) {
       int bmS = tex.getBoundaryModeS();
       int bmT = tex.getBoundaryModeT();
       if ((bmS == Texture.WRAP) && (bmT == Texture.WRAP))
         System.out.println("Texture Boundary Modes both wrap");
       else if ((bmS == Texture.CLAMP) && (bmT == Texture.CLAMP))
         System.out.println("Texture Boundary Modes are both clamped");
       else
         System.out.println("Texture Boundary Modes (s,t): (" +
                              bmS + ", " + bmT + ")");
     }

     // NEW CODE (21/10/04)
     TextureAttributes ta;
     if ((mat != null) && (tex != null)) {
       System.out.println("Setting Material and Texture modulation");
       ta = app.getTextureAttributes();
       if (ta == null)   // no texture attributes found
         ta = new TextureAttributes();
       ta.setTextureMode(TextureAttributes.MODULATE);
       app.setTextureAttributes(ta);
     }

     ta = app.getTextureAttributes();
     if (ta == null)
       System.out.println("No Texture Attributes");
     // else
     //   System.out.println(ta.toString());

  }  // end of printAppearance()


  private void printMaterial(Material mat)
  /*  Write the material's ambient, emissive, diffuse, and specular color 
      details, and the shininess setting, into the model method 
      setMatColours().

     If there are multiple shapes in the model, then a number is included
     at the end of the method name.
  */
  {
    String methodName;
    if (hasMultipleShapes)
      methodName = "setMatColours" + shapeCount + "()";
    else
      methodName = "setMatColours()";

    fileWrite("  private Material " + methodName + "\n" );
    fileWrite("  // set the material's colour and shininess values\n" );
    fileWrite("  {\n" );
    fileWrite("    Material mat = new Material();\n\n" );

    Color3f col = new Color3f();
    mat.getAmbientColor(col);
    fileWrite("    mat.setColor(Material.AMBIENT, 0x00" + 
                          colToHex(col) + ");\n"); // ignore alpha

    mat.getEmissiveColor(col);
    fileWrite("    mat.setColor(Material.EMISSIVE, 0x00" + 
                           colToHex(col) + ");\n"); // ignore alpha

    mat.getDiffuseColor(col);
    fileWrite("    mat.setColor(Material.DIFFUSE, 0xFF" + colToHex(col) + ");\n");

    mat.getSpecularColor(col);
    fileWrite("    mat.setColor(Material.SPECULAR, 0x00" + 
                            colToHex(col) + ");\n"); // ignore alpha

    fileWrite("    mat.setShininess(" + mat.getShininess() + "f);\n\n");

    fileWrite("    return mat;\n" );
    fileWrite("  } // end of " + methodName + "\n\n\n" );
  }  // end of printMaterial()


  private String colToHex(Color3f col)
  /* Convert the RGB color information (0.0f to 1.0f for each component) 
     into a single hexadecimal. */
  {
    int red = (int) Math.round(col.x * 256.0f);
    if (red == 256) red = 255;

    int green = (int) Math.round(col.y * 256.0f);
    if (green == 256) green = 255;

    int blue = (int) Math.round(col.z * 256.0f);
    if (blue == 256) blue = 255;

    // fileWrite("RGB Colour: (" + red + "," + green + "," + blue + ")\n");

    String redStr = Integer.toHexString(red).toUpperCase();
    String greenStr = Integer.toHexString(green).toUpperCase();
    String blueStr = Integer.toHexString(blue).toUpperCase();

    // ensure that the hexadecimal string has two characters
    String rgbStr = (red < 16) ? ("0" + redStr) : redStr;
    rgbStr += (green < 16) ? ("0" + greenStr) : greenStr;
    rgbStr += (blue < 16) ? ("0" + blueStr) : blueStr;

    // fileWrite("RGB Colour: " + rgbStr + "\n");
    return rgbStr;
  }  // end of colToHex()



  // ------------------- file writing utility -----------------

  private void fileWrite(String s)
  { try {
      ofw.write(s);
    }
    catch( IOException ioe )
    { System.err.println("Cannot write '" + s + "' to " + EXAMINE_FN); }
  }  // end of fileWrite()


} // end of WrapObjView class